//@line 51 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\ThumbnailProtocol.js"

"use strict";

const Cu = Components.utils;
const Cc = Components.classes;
const Cr = Components.results;
const Ci = Components.interfaces;

// define the scheme to use - "http://moz-thumb/..."
const SCHEME = "moz-thumb";

// define the image format of our thumbnails
const CONTENT_TYPE = "image/png";

Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

XPCOMUtils.defineLazyServiceGetter(this, "gAppShellService",
                                   "@mozilla.org/appshell/appShellService;1",
                                   "nsIAppShellService");

XPCOMUtils.defineLazyServiceGetter(this, "gCacheService",
                                   "@mozilla.org/network/cache-service;1",
                                   "nsICacheService");

XPCOMUtils.defineLazyGetter(this, "gNetUtil", function () {
  var obj = {};
  Cu.import("resource://gre/modules/NetUtil.jsm", obj);
  return obj.NetUtil;
});

//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\helper.js"

// ----------
// Function: debug
// Logs a given debug message to the console.
//
// Parameters:
//   aMsg - the log message
function debug(aMsg) {
  aMsg = ("Thumbnails: " + aMsg).replace(/\S{80}/g, "$&\n");
  Services.console.logStringMessage(aMsg);
}

// ----------
// Function: trace
// Logs a given debug message to the console and prints the current stack trace.
//
// Parameters:
//   aMsg - the log message
function trace(aMsg) {
  // cut off the first line of the stack trace, because that's just this function.
  let stack = Error().stack.split("\n").slice(1);

  debug("trace: " + aMsg + "\n" + stack.join("\n"));
}

// ----------
// Function: trace
// Returns a new object, containing the merged properties of <aParent> and
// <aObject>.
//
// Parameters:
//   aParent - the object that extends aObject
//   aObject - the object that will be extended
function extend(aParent, aObject) {
  function copy(aSource, aDest) {
    for (let name in aSource) {
      let getter, setter;

      if (getter = aSource.__lookupGetter__(name))
        aDest.__defineGetter__(name, getter);
      else if (setter = aSource.__lookupSetter__(name))
        aDest.__defineSetter__(name, setter);
      else
       aDest[name] = aSource[name];
    }

    return aDest;
  }

  return copy(aObject, copy(aParent, {}));
}

// ----------
// Function: parseUri
// Parses the given URI and returns its parameters.
//
// Parameters:
//   aUri - the URI to parse
function parseUri(aUri) {
  let params = {};
  let query = aUri.spec.split("?")[1] || "";

  query.split("&").forEach(function (param) {
    let [key, value] = param.split("=").map(decodeURIComponent);
    params[key.toLowerCase()] = value;
  });

  return params;
}
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\protocol.js"

// a white 1x1 image that is delivered as a thumbnail for about:blank
const ABOUT_BLANK_URI = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAA" +
                        "BCAYAAAAfFcSJAAAADUlEQVQImWP4////fwAJ+wP9CNHoHgAAAABJ" +
                        "RU5ErkJggg==";

// specifies the flags specific to our protocol implementation
const PROTOCOL_FLAGS = Ci.nsIProtocolHandler.URI_DANGEROUS_TO_LOAD |
                       Ci.nsIProtocolHandler.URI_NORELATIVE |
                       Ci.nsIProtocolHandler.URI_NOAUTH;

// ##########
// Class: Protocol
// Implements the thumbnail protocol handler responsible for moz-thumb:// URIs.
function Protocol() {
}

Protocol.prototype = {
  get scheme() SCHEME,
  get defaultPort() -1,
  get protocolFlags() PROTOCOL_FLAGS,

  // ----------
  // Function: newURI
  // Creates a new URI object that is suitable for loading by this protocol.
  //
  // Parameters:
  //   aSpec - URI string in UTF8 encoding
  //   aOriginCharset - charset of the document from which the URI originated
  newURI: function Proto_newURI(aSpec, aOriginCharset) {
    let uri = Cc["@mozilla.org/network/simple-uri;1"].createInstance(Ci.nsIURI);
    uri.spec = aSpec;
    return uri;
  },

  // ----------
  // Function: newChannel
  // Constructs a new channel from the given URI for this protocol handler.
  //
  // Parameters:
  //   aUri - the URI for which to construct a channel
  newChannel: function Proto_newChannel(aUri) {
    let {url} = parseUri(aUri);

    // don't try to load thumbnails for about:blank pages, just deliver a
    // white 1x1 image. so we make sure that tests can use about:blank
    // and don't have to wait until all the images have been rendered.
    if (0 == url.indexOf("about:blank"))
      return Services.io.newChannel(ABOUT_BLANK_URI, "UTF8", null);

    // prevent recursion if a thumbnail for a thumbnail is requested
    if (0 == url.indexOf(SCHEME + "://"))
      aUri = Services.io.newURI(url, "UTF8", null);

    return new RedirectChannel(aUri);
  },

  // ----------
  // Function: allowPort
  // Returns always false because we'll never want to allow blacklisted ports.
  //
  // Parameters:
  //   aPort - the blacklisted port
  //   aScheme - the scheme the port is blacklisted for
  allowPort: function Proto_allowPort(aPort, aScheme) {
    return false;
  },

  classID: Components.ID("{5a4ae9b5-f475-48ae-9dce-0b4c1d347884}"),
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIProtocolHandler])
};

let NSGetFactory = XPCOMUtils.generateNSGetFactory([Protocol]);
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\baseChannel.js"

let BaseChannel = {
  _wasOpened: false,
  _wasStarted: false,
  _isCanceled: false,

  _setup: function BCH__setup(aUri) {
    this._uri = aUri;

    // nsIChannel
    this.originalURI = aUri;

    // nsIHttpChannel
    this._requestHeaders = {};
    this._responseHeaders = {"content-type": CONTENT_TYPE};
  },

  _startRequest: function CCH__startRequest() {
    try {
      this._listener.onStartRequest(this, this._context);
    } catch (e) {
      // the listener might throw if the request has been canceled
      this.cancel(Cr.NS_BINDING_ABORTED);
    }

    return !this.canceled;
  },

  _serveData: function CCH__serveData(aDataStream) {
    try {
      let available = aDataStream.available();
      this._listener.onDataAvailable(this, this._context, aDataStream, 0, available);
    } catch (e) {
      // the listener might throw if the request has been canceled
      this.cancel(Cr.NS_BINDING_ABORTED);
    }

    return !this.canceled;
  },

  _stopRequest: function CCH__stopRequest() {
    try {
      this._listener.onStopRequest(this, this._context, this.status);
    } catch (e) {
      // this might throw but is generally ignored
    }

    this._cleanup();
  },

  _addToLoadGroup: function CCH__addToLoadGroup() {
    if (this.loadGroup)
      this.loadGroup.addRequest(this, this._context);
  },

  _removeFromLoadGroup: function CCH__removeFromLoadGroup() {
    if (!this.loadGroup)
      return;

    try {
      this.loadGroup.removeRequest(this, this._context, this.status);
    } catch (e) {
      // this might throw and is ignored
    }
  },

  _cleanup: function CCH__cleanup() {
    this._removeFromLoadGroup();
    this.loadGroup = null;

    this._isPending = false;

    delete this._listener;
    delete this._context;
    delete this._entry;
  },

  // nsIChannel
  contentType: CONTENT_TYPE,
  contentLength: -1,
  owner: null,
  contentCharset: null,
  notificationCallbacks: null,

  get URI() this._uri,
  get securityInfo() null,

  asyncOpen: function BCH_asyncOpen(aListener, aContext) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },

  open: function BCH_open() {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },

  // nsIHttpChannel
  redirectionLimit: 10,
  requestMethod: "GET",
  allowPipelining: true,
  referrer: null,

  get requestSucceeded() true,
  get responseStatus() 200,
  get responseStatusText() "OK",

  isNoCacheResponse: function () false,
  isNoStoreResponse: function () false,

  getRequestHeader: function BCH_getRequestHeader(aHeader) {
    aHeader = aHeader.toLowerCase();
    let headers = this._requestHeaders;

    if (aHeader in headers)
      return headers[aHeader];

    throw Cr.NS_ERROR_NOT_AVAILABLE;
  },

  setRequestHeader: function BCH_setRequestHeader(aHeader, aValue, aMerge) {
    if (aMerge)
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;

    this._requestHeaders[aHeader.toLowerCase()] = aValue;
  },

  visitRequestHeaders: function BCH_visitRequestHeaders(aVisitor) {
    this._visitHeaders(this._requestHeaders, aVisitor);
  },

  getResponseHeader: function BCH_getResponseHeader(aHeader) {
    aHeader = aHeader.toLowerCase();
    let headers = this._responseHeaders;

    if (aHeader in headers)
      return headers[aHeader];

    throw Cr.NS_ERROR_NOT_AVAILABLE;
  },

  setResponseHeader: function BCH_setResponseHeader(aHeader, aValue, aMerge) {
    if (aMerge)
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;

    this._responseHeaders[aHeader.toLowerCase()] = aValue;
  },

  visitResponseHeaders: function BCH_visitResponseHeaders(aVisitor) {
    this._visitHeaders(this._responseHeaders, aVisitor);
  },

  _visitHeaders: function BCH__visitHeaders(aHeaders, aVisitor) {
    for (let name in aHeaders) {
      let value = aHeaders[name];

      try {
        aVisitor.visitHeader(name, value);
      } catch (e) {
        // the visitor throws to stop the iteration
        return;
      }
    }
  },

  // nsIRequest
  _status: Cr.NS_OK,
  _isPending: false,

  loadFlags: Ci.nsIRequest.LOAD_NORMAL,
  loadGroup: null,

  get name() this._uri.spec,
  get status() this._status,

  isPending: function () this._isPending,
  resume: function () {},
  suspend: function () {},

  cancel: function BCH_cancel(aStatus) {
    if (this.canceled)
      return;

    this._isCanceled = true;
    this._status = aStatus;
  },

  // nsIHttpChannelInternal
  documentURI: null,

  get canceled() this._isCanceled,

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIChannel,
                                         Ci.nsIHttpChannel,
                                         Ci.nsIHttpChannelInternal,
                                         Ci.nsIRequest])
};
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\redirectChannel.js"

// TODO
function RedirectChannel(aUri) {
  this._setup(aUri);
}

RedirectChannel.prototype = extend(BaseChannel, {
  asyncOpen: function RCH_asyncOpen(aListener, aContext) {
    if (this._wasOpened)
      throw Cr.NS_ERROR_ALREADY_OPENED;

    if (this.canceled)
      return;

    this._listener = aListener;
    this._context = aContext;

    this._isPending = true;
    this._wasOpened = true;

    this._fetchEntry();
  },

  _fetchEntry: function RCH__fetchEntry() {
    let self = this;
    let uri = this._uri;

    Cache.getReadEntry(this.name, function (aEntry) {
      let channel;

      if (aEntry)
        // entry found, deliver from cache
        channel = new CacheChannel(uri, aEntry);
      else
        // entry not found, create a thumbnail and deliver it when ready
        channel = new GenerateChannel(uri);

      self._redirect(channel);
    });
  },

  _redirect: function RCH__redirect(aChannel) {
    let sink = this.notificationCallbacks.QueryInterface(Ci.nsIChannelEventSink);
    let flags = Ci.nsIChannelEventSink.REDIRECT_INTERNAL;
    let callback = this._createRedirectCallback(aChannel);

    sink.asyncOnChannelRedirect(this, aChannel, flags, callback);
  },

  _createRedirectCallback: function RCH__createRedirectCallback(aChannel) {
    let self = this;

    return {
      onRedirectVerifyCallback: function (aStatus) {
        if (!Components.isSuccessCode(aStatus))
          throw Cr.NS_ERROR_UNEXPECTED;

        // cancel the current request before opening the new one
        self.cancel(Cr.NS_OK);

        aChannel.loadGroup = self.loadGroup;
        aChannel.asyncOpen(self._listener, self._context);
      },

      QueryInterface: XPCOMUtils.generateQI([Ci.nsIAsyncVerifyRedirectCallback])
    };
  }
});
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\cacheChannel.js"

// TODO
function CacheChannel(aUri, aEntry) {
  this._entry = aEntry;

  try {
    let contentLength = aEntry.getMetaDataElement("content-length");
    this.setResponseHeader("content-length", contentLength);
  } catch (e) {
    // throws if the meta data element doesn't exist, ignore
  }

  this._setup(aUri);
}

CacheChannel.prototype = extend(BaseChannel, {
  asyncOpen: function CCH_asyncOpen(aListener, aContext) {
    if (this._wasOpened)
      throw Cr.NS_ERROR_ALREADY_OPENED;

    if (this.canceled)
      return;

    this._listener = aListener;
    this._context = aContext;

    this._isPending = true;
    this._wasOpened = true;

    this._serveEntry();
  },

  cancel: function CCH__cancel(aStatus) {
    if (this.canceled)
      return;

    BaseChannel.cancel.call(this, aStatus);

    this._cleanup();
  },

  _serveEntry: function CCH__serveEntry() {
    let self = this;
    let listener = this._listener;
    let context = this._context;

    this._readEntry(function (aDataStream) {
      self._dataStream = aDataStream;

      if (self._startRequest()) {
        self._addToLoadGroup();

        if (self._serveData(aDataStream))
          self._stopRequest();
      }
    });
  },

  _readEntry: function CCH__readEntry(aCallback) {
    let self = this;

    this._readEntryData(function (aDataStream) {
      if (self.canceled)
        return;

      // failed to get the thumbnail data
      if (!aDataStream) {
        self.cancel(Cr.NS_ERROR_UNEXPECTED);
        return;
      }

      aCallback(aDataStream);
    });
  },

  _readEntryData: function CCH__readEntryData(aCallback) {
    let inputStream = this._entry.openInputStream(0);

    gNetUtil.asyncFetch(inputStream, function (aDataStream, aStatus, aRequest) {
      inputStream.close();

      if (!Components.isSuccessCode(aStatus) || !aDataStream.available())
        aDataStream = null;

      aCallback(aDataStream);
    });
  }
});
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\generateChannel.js"

// TODO
function GenerateChannel(aUri) {
  this._setup(aUri);
}

GenerateChannel.prototype = extend(BaseChannel, {
  asyncOpen: function GCH_asyncOpen(aListener, aContext) {
    if (this._wasOpened)
      throw Cr.NS_ERROR_ALREADY_OPENED;

    if (this.canceled)
      return;

    this._listener = aListener;
    this._context = aContext;

    this._isPending = true;
    this._wasOpened = true;

    let self = this;

    Cache.getWriteEntry(this.name, function (aEntry) {
      // we might've been canceled while waiting for the entry
      if (self.canceled)
        return;

      // for some reason we don't have a valid entry
      if (!aEntry) {
        self.cancel(Cr.NS_ERROR_UNEXPECTED);
        return;
      }

      self._createAndServeThumbnail(aEntry);
    });
  },

  _createAndServeThumbnail: function GCH__createAndServeThumbnail(aEntry) {
    if (!this._startRequest())
      return;

    this._addToLoadGroup();

    let self = this;
    let {url, width, height} = parseUri(this._uri);

    // create a new thumbnail
    Thumbnailer.create(url, width, height, function (aDataStream, aStatusCode) {
      // failed to get the thumbnail data
      if (!aDataStream || !aDataStream.available()) {
        self.cancel(Cr.NS_ERROR_UNEXPECTED);
        return;
      }

      // store thumbnails in the cache only for 2xx status codes
      if (Math.floor(aStatusCode / 100) == 2)
        self._prepareCacheEntry(aEntry, aDataStream);

      if (self._serveData(aDataStream))
        self._stopRequest();
    });
  },

  _prepareCacheEntry: function GHC__prepareCacheEntry(aEntry, aDataStream) {
    // create a tee listener to simultaneously deliver data to the
    // stream listener and to the cache entry's stream
    this._useWriteEntryTeeListener(aEntry);

    // store content-length header
    let available = "" + aDataStream.available();
    aEntry.setMetaDataElement("content-length", available);

    // store this entry for two days
    let unixTime = Math.floor(Date.now() / 1000);
    aEntry.setExpirationTime(unixTime + (86400 * 2));
  },

  _useWriteEntryTeeListener: function GCH__useWriteEntryTeeListener(aEntry) {
    let tee = Cc["@mozilla.org/network/stream-listener-tee;1"]
              .createInstance(Ci.nsIStreamListenerTee);

    let outputStream = aEntry.openOutputStream(0);
    tee.init(this._listener, outputStream);

    this._listener = tee;
  }
});
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\cache.js"

// the cache client identifier to separate our cache content from others
const CACHE_CLIENT_ID = "moz-thumb";

// ##########
// Class: Cache
// Singleton providing access to the internal cache.
let Cache = {
  // lazy getter for the cache session
  get _cacheSession() {
    let session = gCacheService.createSession(CACHE_CLIENT_ID,
                                              Ci.nsICache.STORE_ON_DISK, true);

    delete this._cacheSession;
    return this._cacheSession = session;
  },

  // ----------
  // Function: getReadEntry
  // Calls the given <aCallback> with a cache entry opened for reading.
  //
  // Parameters:
  //   aKey - the key identifying the desired cache entry
  //   aCallback - the callback that is called when the cache entry is ready
  getReadEntry: function Cache_getReadEntry(aKey, aCallback) {
    let self = this;

    // try to open the desired cache entry
    this._openCacheEntry(aKey, Ci.nsICache.ACCESS_READ, function (aEntry) {
      if (!self._isValidReadEntry(aEntry))
        aEntry = null;

      aCallback(aEntry);
    });
  },

  // ----------
  // Function: getWriteEntry
  // Calls the given <aCallback> with a cache entry opened for writing.
  //
  // Parameters:
  //   aKey - the key identifying the desired cache entry
  //   aCallback - the callback that is called when the cache entry is ready
  getWriteEntry: function Cache_getWriteEntry(aKey, aCallback) {
    // try to open the desired cache entry
    this._openCacheEntry(aKey, Ci.nsICache.ACCESS_WRITE, aCallback);
  },

  // ----------
  // Function: _isValidReadEntry
  // Returns whether the given cache entry is valid and can be read from.
  //
  // Parameters:
  //   aEntry - the cache entry to verify
  _isValidReadEntry: function Cache__isValidReadEntry(aEntry) {
    if (!aEntry)
      return false;

    let inputStream = aEntry.openInputStream(0);

    // check if the entry actually contains data
    return inputStream && inputStream.available();
  },

  // ----------
  // Function: _openCacheEntry
  // Opens the cache entry identified by the given <aKey>.
  //
  // Parameters:
  //   aKey - the key identifying the desired cache entry
  //   aAccess - the desired access mode
  //   aCallback - the callback to be called when the cache entry was opened
  _openCacheEntry: function Cache__openCacheEntry(aKey, aAccess, aCallback) {
    function onCacheEntryAvailable(aEntry, aAccessGranted, aStatus) {
      let validAccess = (aAccess == aAccessGranted);
      let validStatus = Components.isSuccessCode(aStatus);

      if (aEntry && !(validAccess && validStatus)) {
        aEntry.close();
        aEntry = null;
      }

      aCallback(aEntry);
    }

    let listener = this._createCacheListener(onCacheEntryAvailable);
    this._cacheSession.asyncOpenCacheEntry(aKey, aAccess, listener);
  },

  // ----------
  // Function: _createCacheListener
  // Returns a cache listener implementing the nsICacheListener interface.
  //
  // Parameters:
  //   aCallback - the callback to be called when the cache entry is available
  _createCacheListener: function Cache__createCacheListener(aCallback) {
    return {
      onCacheEntryAvailable: aCallback,
      QueryInterface: XPCOMUtils.generateQI([Ci.nsICacheListener])
    };
  }
};
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\thumbnailer.js"

// the XUL namespace used to create elements in XUL documents
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

// as long as there's bug 565388, we need some lightweight chrome page
// to use as a host frame where we can create some XUL <iframes>
const XUL_FRAME_SRC = "chrome://global/content/mozilla.xhtml";

// the minimum width of the window we're generating thumbnails from
const THUMB_MIN_WIDTH = 1024;

// the color our thumbnails are using for transparent areas
const THUMB_BG_COLOR = "#fff";

// ##########
// Class: Thumbnail
// Creates thumbnails for a given URL by loading it in an <iframe> in a hidden
// window, drawing the window contents to a canvas and finally providing the
// data stream to be read from.
let Thumbnailer = {
  create: function TN_create(aUrl, aWidth, aHeight, aCallback) {
    let frameWidth = Math.max(aWidth, THUMB_MIN_WIDTH);
    let frameHeight = frameWidth / (aWidth / aHeight);

    let self = this;

    HiddenFrame.create(aUrl, frameWidth, frameHeight, function (aFrame, aStatusCode) {
      let window = aFrame.contentWindow;

      // creating the hidden frame takes a while so let's create the canvas in
      // the meantime while we're waiting for the hidden frame to load
      let canvas = self._createCanvas(window, aWidth, aHeight);
      let ctx = canvas.getContext("2d");

      // scale canvas accordingly
      let scaler = aWidth / frameWidth;
      ctx.scale(scaler, scaler);

      // draw to the canvas as soon as the frame is initialized
      ctx.drawWindow(window, 0, 0, frameWidth, frameHeight,
                     THUMB_BG_COLOR, ctx.DRAWWINDOW_DO_NOT_FLUSH);

      // read the image data from the canvas and pass it to the callback
      self._readCanvasData(canvas, function (aDataStream) {
        aCallback(aDataStream, aStatusCode);
        HiddenFrame.destroy(aFrame);
      });
    });
  },

  _createCanvas: function TN__createCanvas(aWindow, aWidth, aHeight) {
    let doc = aWindow.document;
    let canvas = doc.createElement("canvas");
    doc.body.appendChild(canvas);

    canvas.setAttribute("moz-opaque", "");
    canvas.style.width = aWidth + "px";
    canvas.style.height = aHeight + "px";
    canvas.style.visibility = "hidden";
    canvas.width = aWidth;
    canvas.height = aHeight;

    return canvas;
  },

  _readCanvasData: function TN__readCanvasData(aCanvas, aCallback) {
    let dataUri = aCanvas.toDataURL(CONTENT_TYPE, "");
    let uri = Services.io.newURI(dataUri, "UTF8", null);

    gNetUtil.asyncFetch(uri, function (aDataStream, aStatus, aRequest) {
      aCallback(aDataStream);
    });
  }
};

// TODO
let HiddenFrame = {
  _hostFrameCallbacks: [],

  create: function HF_create(aUri, aWidth, aHeight, aCallback) {
    let self = this;

    this._getHostFrame(function (aFrame) {
      let doc = aFrame.contentDocument;
      let iframe = doc.createElementNS(XUL_NS, "iframe");
      iframe.setAttribute("type", "content");
      doc.documentElement.appendChild(iframe);

      iframe.style.width = aWidth + "px";
      iframe.style.height = aHeight + "px";

      // hide the scrollbars
      iframe.style.overflow = "hidden";

      // TODO
      let docShell = iframe.docShell;
      docShell.allowAuth = false;
      docShell.allowWindowControl = false;

      self._whenDocumentLoaded(iframe, function () {
        let statusCode = 200;

        try {
          // if the channel is a nsIHttpChannel get its http status
          let channel = docShell.currentDocumentChannel;
          let httpChannel = channel.QueryInterface(Ci.nsIHttpChannel);
          statusCode = httpChannel.responseStatus;
        } catch (e) {
          // not a http channel, ignore
        }

        aCallback(iframe, statusCode);
      });

      iframe.setAttribute("src", aUri);
    });
  },

  destroy: function HF_destroy(aFrame) {
    aFrame.parentNode.removeChild(aFrame);
  },

  _getHostFrame: function HF__getHostFrame(aCallback) {
    // return the cached value if any
    if (this._hostFrame) {
      aCallback(this._hostFrame);
      return;
    }

    // if we're not the first item in the queue we don't need to do a thing
    // but wait for the host frame to be loaded
    if (this._hostFrameCallbacks.push(aCallback) > 1)
      return;

    let window = gAppShellService.hiddenDOMWindow;
    let doc = window.document;

    // create the host frame
    let hostFrame = doc.createElement("iframe");
    doc.documentElement.appendChild(hostFrame);

    // TODO
    hostFrame.setAttribute("src", XUL_FRAME_SRC);

    let self = this;

    this._whenContentLoaded(hostFrame, function () {
      self._hostFrame = hostFrame;
      self._hostFrameCallbacks.forEach(function (callback) callback(hostFrame));
      self._hostFrameCallbacks = null;
    });
  },

  _whenContentLoaded: function HF__whenContentLoaded(aElement, aCallback) {
    aElement.addEventListener("DOMContentLoaded", function onLoad() {
      aElement.removeEventListener("DOMContentLoaded", onLoad, false);
      aCallback();
    }, false);
  },

  _whenDocumentLoaded: function HF__whenDocumentLoaded(aIframe, aCallback) {
    aIframe.addEventListener("load", function onLoad(aEvent) {
      if (aIframe.contentDocument == aEvent.target) {
        aIframe.removeEventListener("load", onLoad, true);
        aCallback();
      }
    }, true);
  }
};
